;;; *************************************************************************
;;;; FILE IDENTIFICATION
;;;;
;;;; Name:          package.lisp
;;;; Purpose:       Audioscrobbler/lastfm wrapper
;;;; Programmer:    Nicolas Lamirault <nicolas.lamirault@gmail.com>
;;;;
;;;; This file, part of audioscrobbler, is Copyright (c) 2006 by Nicolas Lamirault
;;;;
;;;; audioscrobbler users are granted the rights to distribute and use this software
;;;; as governed by the terms of the Lisp Lesser GNU Public License
;;;; (http://opensource.franz.com/preamble.html), also known as the LLGPL.
;;;;
;;;; *************************************************************************


(in-package :audioscrobbler)


;; ------
;; Tools
;; ------


(defun get-current-date ()
 "Get current date in YYYY-MM-DD mm:hh:ss format."
 (multiple-value-bind
       (second minute hour date month year day-of-week dst-p tz)
     (get-decoded-time)
   (declare (ignore day-of-week dst-p tz))
   (format nil "~2,'0d-~2,'0d-~2,'0d ~2,'0d:~2,'0d:~2,'0d"
           year month date hour minute second)))



(defun get-xml-answer (query)
 "Retreive informations from audioscrobbler and parse XML output."
;;   (multiple-value-bind (body code headers uri kept-alive-state)
;;       (s-http-client:do-http-request (cl-ppcre:regex-replace-all " " query "+"))
;;     (declare (ignore headers uri kept-alive-state))
;;     (when (= code 200)
 (multiple-value-bind (body status-code headers uri stream must-close)
     (drakma:http-request query)
   (declare (ignore headers uri stream must-close))
   (when (and status-code (= status-code 200))
     (s-xml:parse-xml-string body))))


(defun retreive-user-profil-information (username)
 "Information about the profile (eg: Name, Gender)."
 (get-xml-answer (format nil +user-profile-information+ username)))


(defun retreive-user-top-artists (username)
 "50 most played artists from a music profile."
 (get-xml-answer (format nil +user-top-artists+  username)))


(defun retreive-user-top-albums (username)
 "50 most played albums from a music profile."
 (get-xml-answer (format nil +user-top-albums+ username)))


(defun retreive-user-top-tracks (username)
 "50 most played tracks from a music profile."
 (get-xml-answer (format nil +user-top-tracks+ username)))


(defun retreive-user-top-tags (username)
 "50 most played tracks from a music profile."
 (get-xml-answer (format nil +user-top-tags+ username)))


(defun retreive-user-top-tags-artist (username artist-name)
 "Tags most used on a specific artist."
 (get-xml-answer (format nil +user-top-tags-artist+ username artist-name)))


(defun retreive-user-top-tags-album (username artist-name album-name)
 "Tags most used on a specific album."
 (get-xml-answer
  (format nil +user-top-tags-album+ username artist-name album-name)))


(defun retreive-user-top-tags-track (username artist-name track-name)
 "Tags most used on a specific track."
 (get-xml-answer
  (format nil +user-top-tags-track+ username artist-name track-name)))


(defun retreive-user-friends (username)
 "Friends added to this profile."
 (get-xml-answer (format nil +user-friends+ username)))


(defun retreive-user-neighbours (username)
 "People with similar taste to this profile."
 (get-xml-answer (format nil +user-neighbours+ username)))


(defun retreive-user-recent-tracks (username)
 "10 recently played tracks for this profile."
 (get-xml-answer (format nil +user-recent-tracks+ username)))


(defun retreive-user-weekly-artists (username)
 "Most recent weekly artist chart."
 (get-xml-answer (format nil +user-weekly-artist+ username)))


(defun retreive-user-weekly-albums (username)
 "Most recent weekly album chart."
 (get-xml-answer (format nil +user-weekly-album+ username)))


(defun retreive-user-weekly-tracks (username)
 "Most recent weekly track chart."
 (get-xml-answer (format nil +user-weekly-track+ username)))


(defun retreive-artist-related (artist)
 "People who listen to ARTIST-NAME also listen to ..."
 (get-xml-answer (format nil +artist-related+ artist)))


(defun retreive-artist-top-albums (artist)
 "The most popular albums by a given ARTIST-NAME."
 (get-xml-answer (format nil +artist-top-albums+ artist)))


(defun retreive-artist-top-tracks (artist)
 "The most popular tracks by a given ARTIST."
 (get-xml-answer (format nil +artist-top-tracks+ artist)))




;; ----------------
;; API WebServices
;; ----------------


(defclass user ()
 ((name :accessor user-name
        :initarg :name
        :initform nil)
  (url :accessor user-url
       :initarg :url
       :initform nil))
 (:documentation "An user."))


(defclass profil (user)
 ((id :accessor profil-id
      :initarg :id
      :initform nil)
  (registered :accessor profil-registered
              :initarg :registered
              :initform nil)
  (age :accessor profil-age
       :initarg :age
       :initform nil)
  (gender :accessor profil-gender
          :initarg :gender
          :initform nil)
  (country :accessor profil-country
           :initarg :country
           :initform nil))
 (:documentation "Informations about the user."))


(defun get-profil (username)
 "Create a user profil from USERNAME."
 (let ((xmldoc (retreive-user-profil-information username)))
   (make-instance 'profil
                  :id (third (car xmldoc))
                  :name username
                  :url (cadr (second xmldoc))
                  :registered (cadr (fifth xmldoc))
                  :age (cadr (sixth xmldoc))
                  :gender (cadr (seventh xmldoc))
                  :country (cadr (eighth xmldoc)))))


(defgeneric describe-profil (profil &optional stream)
 (:documentation "Print to STREAM PROFIL's informations."))


(defmethod describe-profil ((profil profil) &optional (stream *standard-output*))
 (with-slots (name url id registered age gender country) profil
   (format stream "~A (~A) [~A] ~A ~A ~A ~A~%"
           name gender age country registered id url)))


(defclass content ()
 ((name :accessor content-name
        :initarg :name
        :initform nil)
  (mbid :accessor content-mbid
        :initarg :mbid
        :initform nil)
  (url :accessor content-url
       :initarg :url
       :initform nil))
 (:documentation "A music content."))


(defclass top-content (content)
 ((playcount :accessor top-content-playcount
             :initarg :playcount
             :initform nil)
  (rank :accessor top-content-rank
        :initarg :rank
        :initform nil))
 (:documentation "A most played content."))


(defclass artist (content)
 ()
 (:documentation "A music artist."))


(defclass user-top-artist (top-content)
 ()
 (:documentation "A most played music artist."))


(defgeneric describe-top-artist (user-top-artist &optional stream)
 (:documentation "Print informations about top ARTIST."))


(defmethod describe-top-artist ((user-top-artist user-top-artist)
                               &optional (stream *standard-output*))
 (with-slots (name mbid url) user-top-artist
   (format stream "~A ~A ~A~%" name mbid url)))



(defun get-user-top-artists (username)
 "Retreive a list of most played artists by an user."
 (let ((xmldoc (retreive-user-top-artists username)))
   (when (and (listp (cdr xmldoc))
              (listp (cadr xmldoc)))
     (loop for sexp in (cdr xmldoc)
        as name = (cadr (first (cdr sexp)))
        as mbid = (if (listp (second (cdr sexp)))
                      (cadr (second (cdr sexp)))
                      "")
        as playcount = (cadr (third (cdr sexp)))
        as rank = (cadr (fourth (cdr sexp)))
        as url = (cadr (fifth (cdr sexp)))
        collect (make-instance 'user-top-artist
                               :name name
                               :mbid mbid
                               :url url
                               :playcount playcount
                               :rank rank)))))


(defclass album (content)
 ()
 (:documentation "Album of a music artist."))


(defclass user-top-album (album top-content)
 ((artist-name :accessor :artist-name
               :initarg :artist-name
               :initform nil)
  (large :accessor top-album-large
         :initarg :large
         :initform nil)
  (medium :accessor top-album-medium
         :initarg :medium
         :initform nil)
  (small :accessor top-album-small
         :initarg :small
         :initform nil))
 (:documentation "A most played music album."))



(defun get-user-top-albums (username)
 "Retreive a list of most played music albums by an user."
 (let ((xmldoc (retreive-user-top-albums username)))
   (when (and (listp (cdr xmldoc))
              (listp (cadr xmldoc)))
     (loop for sexp in (cdr xmldoc)
        as artist-name = (cadr (first (cdr sexp)))
        as name = (cadr (second (cdr sexp)))
        as mbid = (if (listp (third (cdr sexp)))
                      (cadr (third (cdr sexp)))
                      "")
        as playcount = (cadr (fourth (cdr sexp)))
        as rank = (cadr (fifth (cdr sexp)))
        as url = (cadr (sixth (cdr sexp)))
        as large = (cadr (first (cdr (seventh (cdr sexp)))))
        as medium = (cadr (second (cdr (seventh (cdr sexp)))))
        as small = (cadr (third (cdr (seventh (cdr sexp)))))
        collect (make-instance 'user-top-album
                               :name name
                               :artist-name artist-name
                               :mbid mbid
                               :url url
                               :playcount playcount
                               :rank rank
                               :small small :medium medium :large large)))))


(defclass user-top-track (top-content)
 ((artist-name :accessor :artist-name
               :initarg :artist-name
               :initform nil))
 (:documentation "A most played music song."))


(defun get-user-top-tracks (username)
 "Retreive a list of most played music tracks by an user."
 (let ((xmldoc (retreive-user-top-tracks username)))
   (when (and (listp (cdr xmldoc))
              (listp (cadr xmldoc)))
     (loop for sexp in (cdr xmldoc)
        as artist = (car (last (first (cdr sexp))))
        as name = (cadr (second (cdr sexp)))
        as playcount = (cadr (fourth (cdr sexp)))
        as rank = (cadr (fifth (cdr sexp)))
        as url = (cadr (sixth (cdr sexp)))
        collect (make-instance 'user-top-track
                               :artist-name artist
                               :name name
                               :playcount playcount
                               :rank rank
                               :url url)))))

;; (defun get-recent-tracks (username)
;;   "10 recently played tracks for this profile."
;;   (let ((xmldoc (retreive-user-recent-tracks username)))



(defclass neighbour (user)
 ((image :accessor neighbours-image
         :initarg :image
         :initform nil)
  (match :accessor neighbours-match
         :initarg :match
         :initform nil))
 (:documentation "People similar to a profile."))


(defun get-user-neighbours (username)
 "People with similar taste to this profile."
 (let ((xmldoc (retreive-user-neighbours username)))
   (loop for sexp in (cdr xmldoc)
      as name = (third (first sexp))
      as url = (cadr (second sexp))
      as image = (cadr (third sexp))
      as match = (cadr (fourth sexp))
      collect (make-instance 'neighbour
                             :name name
                             :url url
                             :image image
                             :match match))))

(defclass friend (user)
 ((image :accessor friends-image
         :initarg :image
         :initform nil)
  (connection :accessor friends-connection
              :initarg :connection
              :initform nil))
 (:documentation "Profile's friend."))


(defun get-user-friends (username)
 "Friends added to this profile."
 (let ((xmldoc (retreive-user-friends username)))
   (when (and (listp (cdr xmldoc))
              (listp (cadr xmldoc)))
     (loop for sexp in (cdr xmldoc)
        as name = (third (first sexp))
        as url = (cadr (second sexp))
        as image = (cadr (third sexp))
        as connection = (cadadr (fourth sexp))
        collect (make-instance 'friend
                               :name name
                               :url url
                               :image image
                               :connection connection)))))


(defclass artist-related (artist)
 ((image-small :accessor artist-related-image-small
               :initarg :image-small
               :initform nil)
  (image :accessor artist-related-image
         :initarg :image
         :initform nil))
 (:documentation "A related artist."))


(defun get-artist-related (artist)
 "People who listen to ARTIST also listen to A,B,C..."
 (let ((xmldoc (retreive-artist-related artist)))
   (when (and (listp (cdr xmldoc))
              (listp (cadr xmldoc)))
     (loop for sexp in (cdr xmldoc)
        as name = (cadr (first (cdr sexp)))
        as mbid = (if (listp (second (cdr sexp)))
                      (cadr (second (cdr sexp)))
                      "")
        as url = (cadr (third (cdr sexp)))
        as image-small = (cadr (fourth (cdr sexp)))
        as image = (cadr (fifth (cdr sexp)))
        collect (make-instance 'artist-related
                               :name name
                               :mbid mbid
                               :url url
                               :image-small image-small
                               :image image)))))


(defclass artist-top-album (content)
 ((reach :accessor artist-top-album-reach
         :initarg :reach
         :initform nil))
 (:documentation "A popular artist's album."))


(defun get-artist-top-albums (artist)
 "The most popular albums by a given ARTIST."
 (let ((xmldoc (retreive-artist-top-albums artist)))
   (when (and (listp (cdr xmldoc))
              (listp (cadr xmldoc)))
     (loop for sexp in (cdr xmldoc)
          as name = (cadr (first (cdr sexp)))
          as mbid = (if (listp (second (cdr sexp)))
                      (cadr (second (cdr sexp)))
                      "")
          as reach = (cadr (third (cdr sexp)))
          as url = (cadr (fourth (cdr sexp)))
          collect (make-instance 'artist-top-album
                                 :name name
                                 :mbid mbid
                                 :reach reach
                                 :url url)))))


(defclass artist-top-track (content)
 ((reach :accessor artist-top-track-reach
         :initarg :reach
         :initform nil))
 (:documentation "A most played music song for an artist."))


(defun get-artist-top-tracks (artist)
 "The most popular tracks by a given ARTIST."
 (let ((xmldoc (retreive-artist-top-albums artist)))
   (when (and (listp (cdr xmldoc))
              (listp (cadr xmldoc)))
     (loop for sexp in (cdr xmldoc)
          as name = (cadr (first (cdr sexp)))
          as mbid = (if (listp (second (cdr sexp)))
                      (cadr (second (cdr sexp)))
                      "")
          as reach = (cadr (third (cdr sexp)))
          as url = (cadr (fourth (cdr sexp)))
          collect (make-instance 'artist-top-track
                                 :name name
                                 :mbid mbid
                                 :reach reach
                                 :url url)))))


;; ---------------------
;; HTTP IO Protocol 1.1
;; ---------------------


(unless (boundp '+client-id+)
  (defconstant +client-id+ "tst"))

(unless (boundp '+client-version+)
  (defconstant +client-version+ "0.1"))


(unless (boundp '+protocol-identify+)
 (defconstant +protocol-identify+
   "http://post.audioscrobbler.com/?hs=true&p=1.1&c=~A&v=~A&u=~A"))


(unless (boundp '+protocol-post+)
 (defconstant +protocol-post+
   "a%5b0%5d=~A&t%5b0%5d=~A&b%5b0%5d=~A&m%5b0%5d=~A&l%5b0%5d=~A&i%5b0%5d=~A&u=~A&"))



(defun get-md5sum-code (password challenge)
 "Create a string like md5(md5(PASSWORD) + CHALLENGE)."
 (map 'string #'code-char
      (md5:md5sum-sequence
       (concatenate 'string
                    (map 'string #'code-char (md5:md5sum-sequence password))
                    challenge))))


(defun get-identification (username &optional (client-id +client-id+)
                          (client-version +client-version+))
 "Make an HTTP request to retreive challenge for protocol identification.
Returns 2 values : ID and URL."
 (let ((query (format nil +protocol-identify+
                      client-id client-version username)))
;;     (multiple-value-bind (body code headers uri kept-alive-state)
;;         (s-http-client:do-http-request query :method :post)
;;       (declare (ignore headers uri kept-alive-state))
;;       (when (= code 200)
   (multiple-value-bind (body status-code headers uri stream must-close)
       (drakma:http-request query)
     (declare (ignore headers uri stream must-close))
     (when (and status-code (= status-code 200))
       (with-input-from-string (stream body)
         (loop as line = (read-line stream nil)
            until (null line)
            when (cl-ppcre:scan "UPDATE" line)
            return (let ((challenge (read-line stream nil))
                         (url (read-line stream nil)))
                     (values challenge url))))))))


(defun post-song (username password artist album song length time)
 (multiple-value-bind (challenge url)
     (get-identification username)
   (let* ((md5 (get-md5sum-code password challenge))
          (query (format nil (concatenate 'string url "?" +protocol-post+)
                         artist song album
                         length time (get-current-date) username))
          (uri (cl-ppcre:regex-replace-all " "
                                           (url-rewrite:add-get-param-to-url query "s" md5)
                                           "+")))
     (print uri)
;;       (multiple-value-bind (body code headers uri kept-alive-state)
;;           (s-http-client:do-http-request uri)
;;         (declare (ignore headers uri kept-alive-state))
;;         (when (= code 200)
     (multiple-value-bind (body status-code headers uri stream must-close)
       (drakma:http-request query)
     (declare (ignore headers uri stream must-close))
     (when (and status-code (= status-code 200))
         (with-input-from-string (stream body)
           (loop as line = (read-line stream nil)
              until (null line)
              do (print line))))))))
